# External libraries
from z3 import *             
import copy

from planet.design import Design
from planet.constraint import (
    Counterbalance, NoRepeat, InnerBlock, OuterBlock,
    AbsoluteRank, Order
)
from planet.region import DesignRegion


def nest_structure(d1, d2):
    constraints = []
    # Match all variables from the outer design within each block matrix
    
    # NOTE: changed this from design variables to variables. Check that nothing breaks.
    for variable in d2.variables:
        constraints.append(InnerBlock(
            variable,
            DesignRegion(d1.get_width(),d1.num_plans(),[1, 1])
        ))

     # Match all variables from the inner design across every block
    for variable in d1.variables:
        constraints.append(OuterBlock(
            variable,
            d1.get_width(),
            d1.num_plans(),
            stride = [1, 1]
        ))

    return constraints


def copy_nested_constraints(design1, design2):
    width1, width2 = design1.get_width(), design2.get_width()
    total_groups = design1.num_plans() * design2.num_plans()
    total_conditions = width1 * width2
    constraints = []
    # need to modify the inner constraints region
    # Add counterbalance constraints from design1
    for constraint in design1.constraints.get_constraints():
        if isinstance(constraint, Counterbalance):
            constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=constraint.width or design1.get_width(),
                    height=constraint.height or design1.num_plans(),
                    stride=constraint.stride
                )
            )

        elif isinstance(constraint, NoRepeat):
            constraints.append(
                NoRepeat(
                    constraint.variable,
                    width=constraint.width if constraint.width < design1.get_width() else design1.get_width(),
                    stride=constraint.stride
                )
            )

        elif isinstance(constraint, InnerBlock):
            
            constraints.append(
                InnerBlock(
                    constraint.variable,
                    DesignRegion(
                        constraint.width or design1.get_width(),
                        constraint.height,
                        constraint.stride
                    )
                )
            )

        elif isinstance(constraint, AbsoluteRank):
            # Modify the existing constraint's width and stride
            new_constraint = copy.copy(constraint)
            new_constraint.width = constraint.width if constraint.width < design1.get_width() else design1.get_width()
            new_constraint.stride = constraint.stride
            constraints.append(new_constraint)

        elif isinstance(constraint, Order):
            # Modify the existing constraint's width and stride
            new_constraint = copy.copy(constraint)
            new_constraint.width = constraint.width if constraint.width < design1.get_width() else design1.get_width()
            new_constraint.stride = constraint.stride
            constraints.append(new_constraint)

        else:
            constraints.append(copy.copy(constraint))


    # need to modify out constraint region
    for constraint in design2.constraints.get_constraints():
        # FIXME FIXME, I think this is a ratio problem
        if isinstance(constraint, Counterbalance):
            stride = [design1.num_plans(), width1 * constraint.stride[1]]
            # Add counterbalance constraint for design2 variables
            constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=total_conditions,
                    height=total_groups,
                    stride=stride
                )
            )

        elif isinstance(constraint, NoRepeat):
            width = constraint.width if constraint.width < width2 else width2
            constraints.append(
                NoRepeat(
                    constraint.variable,
                    width*width1,
                    constraint.stride*width1
                )
            )
        
        elif isinstance(constraint, AbsoluteRank):
            # Modify the existing constraint's width and stride
            new_constraint = copy.copy(constraint)
            new_constraint.width = constraint.width if constraint.width < width2 else width2
            new_constraint.width *= width1
            new_constraint.stride *= width1
            constraints.append(new_constraint)

        elif isinstance(constraint, Order):
            # Modify the existing constraint's width and stride
            new_constraint = copy.copy(constraint)
            new_constraint.width = constraint.width if constraint.width < width2 else width2
            new_constraint.width *= width1
            new_constraint.stride *= width1
            constraints.append(new_constraint)
    
        elif isinstance(constraint, InnerBlock):
            constraints.append(
                InnerBlock(
                    constraint.variable, 
                    DesignRegion(
                    constraint.width*width1, 
                    constraint.height*design1.num_plans(), 
                    [1, 1]
                    )
                )
            )

        # # here I need to multiply stride by the number of conditions of the block variable
        elif isinstance(constraint, OuterBlock):
            constraints.append(
                OuterBlock(
                    constraint.variable, 
                    constraint.width*width1, 
                    constraint.height*design1.num_plans(), 
                    stride = [1, 1])
            )
            
        else:
            constraints.append(copy.copy(constraint))
            
    return constraints


def add_design_variables(*, des, combined_des):
    for v in des.design_variables:
        combined_des.add_variable(v)



def nest(*, outer:Design, inner:Design):
    """
    Nest two designs to create a combined experimental design.

    Args:
        outer: The outer design object.
        inner: The inner design object.

    Returns:
        A new Design object representing the nested experimental design.
    """
    total_groups = inner.num_plans() * outer.num_plans()
    total_conditions = inner.get_width() * outer.get_width()


    combined_design = (
        Design()
        .limit_plans(total_groups)
        .num_trials(total_conditions)
    )

    add_design_variables(des = inner, combined_des = combined_design)
    add_design_variables(des = outer, combined_des = combined_design)
    # Nest structural constraints and copy semantic ones
    copied_constraints = copy_nested_constraints(inner, outer)
    if copied_constraints:
        combined_design.add_constraints(copied_constraints)

    combined_design.add_constraints(nest_structure(inner, outer))
    combined_design.set_minimum_trials(inner.get_width() * outer.get_width())

    return combined_design







from z3 import *
from .candl import *

class BitVectors:
    def __init__(self, n, variables):
        self.variables = variables
        self.len = self.determine_num_bits()
        self.z3_variables =  [BitVec(f"C_{index + 1}", self.len) for index in range(n)]

    def get_possible_values(self):
        values = []
        k = 0
        for i in range(len(self.variables)):
            variable = self.variables[i]
            n = variable.n

            if not i:
                values.extend([x for x in range(n)])
            else:
                k += int(math.ceil(math.log(self.variables[i-1].n, 2))) + 1

                new_values = []
                for v in values:
                    for y in range(n):
                        x = y << k 
             
                        new_values.append(x + v)
                
                values.extend(new_values)

            if not len(values):
                values.extend([x for x in range(n)])

        return set(values)
    
    def determine_num_bits_total(self):
        bitvec_length = 0
        for variable in self.variables:
            bits = int(math.ceil(math.log(variable.n, 2)))
            bitvec_length += bits

        # necessary for two's compliment bit
        bitvec_length += len(self.variables)
        return bitvec_length
    
    def determine_num_variable_bits(self, variable):
        return int(math.ceil(math.log(variable.n, 2)))

    def determine_num_bits(self, variable=None):

        if variable is None:
            return self.determine_num_bits_total()
        else:
            return self.determine_num_variable_bits(variable)

    

    def get_lower_bits(self, var):
        var_index = self.variables.index(var)
        bit_index = 0
        # for all the variables already represented
        for i in range(var_index):
            # length of all of the other variable bit representations
            bit_index += get_num_bits(len(self.variables[i]))

        # accounts for two's compliment bit
        bit_index += var_index
        return bit_index
    
    def get_variables(self):
        return self.z3_variables
    
    def get_variable_assignment(self, var, z3):
        lo = self.get_lower_bits(var)
        length = get_num_bits(len(var))
        return Extract(lo+length, lo, z3)

    
    def get_variable_assignments(self, var, z3s):
        assignments = []

        for z3 in z3s:
            assignments.append(self.get_variable_assignment(var, z3))

        return assignments
    
    def new_bitvector(self):
        # FIXME: hacky
        return BitVec("x", self.len)
    
    def int_as_bitvec(self, val, size):
        return BitVecVal(val, size)
    

class TupleVariables:
    def __init__(self, n, variables):
        """
        Initialize the TupleVariables class.

        Args:
            n (int): The number of Z3 tuple variables to create.
            variables (list): A list of ExperimentVariable objects.
        """
        self.variables = variables
        self.tuple_sort, self.tuple_ctor, self.tuple_proj = TupleSort("IntTuple", [IntSort() for _ in variables])
        self.z3_variables = [Const(f"C_{index + 1}", self.tuple_sort) for index in range(n)]

    def get_possible_values(self):
        """
        Get all possible values for the tuple variables.

        Returns:
            set: A set of Z3 tuples representing all possible values.
        """
        values = []
        for variable in self.variables:
            n = variable.n
            values.append([x for x in range(n)])
        return set(product(*values))  # Cartesian product of all variable values

    def get_variables(self):
        """
        Get the Z3 tuple variables.

        Returns:
            list: A list of Z3 tuple variables.
        """
        return self.z3_variables
    
    def tuple_to_int(self, t):
        n = len(self.tuple_proj)
        return Sum([
            self.tuple_proj[i](t) * (10 ** (n - i - 1))
            for i in range(n)
        ])
    

    def get_variable_assignment(self, var, z3_tuple):
        """
        Extract the assignment for a specific variable from a Z3 tuple.

        Args:
            var (ExperimentVariable): The variable to extract.
            z3_tuple (Z3 tuple): The Z3 tuple variable.

        Returns:
            z3.Int: The Z3 integer representing the variable's assignment.
        """
        var_index = self.variables.index(var)
        return self.tuple_proj[var_index](z3_tuple)

    def get_variable_assignments(self, var, z3_tuples):
        """
        Extract the assignments for a specific variable from a list of Z3 tuples.

        Args:
            var (ExperimentVariable): The variable to extract.
            z3_tuples (list): A list of Z3 tuple variables.

        Returns:
            list: A list of Z3 integers representing the variable's assignments.
        """
        return [self.get_variable_assignment(var, z3_tuple) for z3_tuple in z3_tuples]



    # def new_tuple_variable(self, name):
    #     """
    #     Create a new Z3 tuple variable.

    #     Args:
    #         name (str): The name of the Z3 tuple variable.

    #     Returns:
    #         Z3 tuple: A new Z3 tuple variable.
    #     """
    #     return self.create_z3_tuple(name)

    # def int_as_tuple(self, values):
    #     """
    #     Convert a list of integers into a Z3 tuple constant.

    #     Args:
    #         values (list): A list of integer values.

    #     Returns:
    #         Z3 tuple: A Z3 tuple constant.
    #     """
    #     return self.tuple_sort.constructor(*[IntVal(value) for value in values])


# External libraries
from z3 import *             
import copy
from planet.design import Design
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat, InnerBlock, OuterBlock,
    SetRank, SetPosition, AbsoluteRank
)
from planet.candl import combine_lists
from planet.region import DesignRegion


def cross_structure(d1, d2):
    constraints = []
    # Match all variables from the outer design within each block matrix
    for i in range(len(d2.variables)):
        constraints.append(InnerBlock(
            d2.variables[i],
            DesignRegion(1, d1.num_plans(), [1, 1])
        ))

     # Match all variables from the inner design across every block
    for i in range(len(d1.variables)):
        constraints.append(OuterBlock(
            d1.variables[i],
            d1.get_width(),
            d1.num_plans(),
            stride = [1, 1]
        ))
    return constraints


def copy_crossed_constraints(design1, design2, total_conditions, total_groups):
    constraints = []
     # need to modify the inner constraints region
    # Add counterbalance constraints from design1
    for constraint in design1.constraints.constraints:
        if isinstance(constraint, Counterbalance):
            if constraint.width and constraint.height:
                constraints.append(
                    Counterbalance(
                        constraint.variable,
                        width=constraint.width,
                        height=constraint.height,
                        stride=constraint.stride
                    )
                )
            else: 
                constraints.append(
                    Counterbalance(
                        constraint.variable,
                        width=design1.get_width(),
                        height=design1.num_plans(),
                        stride=constraint.stride
                    )
                )

        else:
            constraints.append(
                    copy.copy(constraint)
                )

    # need to modify out constraint region
    for constraint in design2.constraints.constraints:
        if isinstance(constraint, Counterbalance):
            stride_height = constraint.height if constraint.height else design1.num_plans()
            # Add counterbalance constraint for design2 variables
            constraints.append(
                Counterbalance(
                    constraint.variable,
                    width=total_conditions,
                    height=total_groups,
                    stride=[stride_height, 1]
                )
            )
    
        elif isinstance(constraint, InnerBlock):
            stride_height = constraint.height if constraint.height else design1.num_plans()
            constraints.append(
                InnerBlock(constraint.variable, DesignRegion(constraint.width,constraint.height*design1.num_plans(), [1, 1]))
            )

        # # here I need to multiply stride by the number of conditions of the block variable
        elif isinstance(constraint, OuterBlock):
            stride_height = constraint.height if constraint.height else design1.num_plans()
            constraints.append(
                OuterBlock(constraint.variable, constraint.width, constraint.height*design1.num_plans(), stride = [stride_height, 1])
            )

        else:
            constraints.append(
                    copy.copy(constraint)
                )

    return constraints



# NOTE: this is an absolute mess :(
# FIXME: come back to this. Won't generalize...
# need to fix how the match blocks work, but this is not a priority
def cross(design1, design2):
    """
    Nest two designs together to create a combined experimental design.
    
    Args:
        design1: First design object
        design2: Second design object
        
    Returns:
        Combined design object
    """

 
    # Calculate the total number of groups in the combined design
    total_groups = design1.num_plans() * design2.num_plans()
    # Combine variables from both designs
    combined_variables = combine_lists(design1.variables, design2.variables)
    # Calculate width1 (the product of all variable lengths in design1)
    width1 = design1.get_width()
    width2 = design2.get_width()
    
    assert width1 == width2
    total_conditions = width2 
    
    # Create a new design with the combined variables
    combined_design = ( Design()
                       .limit_plans(total_groups)
                       .num_trials(total_conditions)
                    )
    
    combined_design.variables.extend(combined_variables)
    combined_design.add_constraints(copy_crossed_constraints(design1, design2, total_conditions, total_groups))
    combined_design.add_constraints(cross_structure(design1, design2))
    combined_design.set_minimum_trials(design1.get_width())

    return combined_design


from planet.candl import generate_conditions
from planet.constraint import InnerBlock, OuterBlock
from planet.variable import MultiFactVariable



class Randomizer:
    def __init__(self, variable, width, span, n):
        self.rand_vars = self._generate_random_variables(n, variable, width)
        self.n = n
        self.variable = variable
        self.width = width
        self.span = span

      
    def _determine_random_width(self, rand):
        width = self.get_width()
        div = 1
        for constraint in self.constraints.get_constraints_for_variable(rand):
            if isinstance(constraint, OuterBlock):
                width = constraint.width
  
            elif isinstance(constraint, InnerBlock):
                div = constraint.width
        
        return int(width/div)
    

    def _determine_random_span(self, rand):
        span = 1
        for constraint in self.constraints.get_constraints_for_variable(rand):
            if isinstance(constraint, InnerBlock):
               span = constraint.width
        return span
    
    def _instantiate_random_variables(self, n, rand, plans):
        # NOTE to self: this will only work if there is one random variable :)
        """
        Think about this like instantiating the elements of a matrix of random variables
        """
        assert plans is not None
        width = self._determine_random_width(rand)
        span = self._determine_random_span(rand)
        variables = rand.variables if isinstance(rand, MultiFactVariable) else [rand]

        random_index = self.variables.index(variables[0])
        randomizer = Randomizer(rand, width, span, int(n*self.get_width()/width/span))
        return randomizer.apply_randomization(width, span, random_index, n, plans)
     




    def _generate_random_variables(self, n, variable, width):
        ws_conditions = generate_conditions(n, variable, width)
        return ws_conditions
    
    def get_blocks(self, plan, width, span):
        """Chunk the plan into blocks of size width * span."""
        block_size = width * span
        return [plan[i:i + block_size] for i in range(0, len(plan), block_size)]

    def replace_condition(self, condition_str, replacement, index_to_replace):
        """Replace part of a condition string at the given index."""
        parts = condition_str.split("-")
        parts[index_to_replace] = replacement
        return "-".join(parts)
    
 

    def apply_randomization(self, width, span, random_index, n, plans):
        new_plans = []
        for plan_idx, plan in enumerate(plans):
            blocks = self.get_blocks(plan, width, span)
            reps_per_plan = int(n / len(plans))
            for rep_idx in range(reps_per_plan):
                new_plan = []

                for block_idx, block in enumerate(blocks):
                    # FIXME: does  not work when random variables are the outer design
               
                    rand_idx = rep_idx * len(blocks) + plan_idx * int(n/len(plans)) + block_idx
                    rand_var = self.rand_vars[rand_idx]
                    for cond_idx in range(width):
                        for within_block_idx in range(span):
                            index = cond_idx * span + within_block_idx
                            old_condition = block[index]

                            if isinstance(self.variable, MultiFactVariable):
                                variables = self.variable.variables
                                replacement = rand_var[cond_idx]
                                replacement = replacement.split("-")

                                #FIXME: ugly
                                for i in range(len(variables)):
                                    var_replacement = replacement[i]
                                    new_condition = self.replace_condition(old_condition, var_replacement, random_index+i)
                                    old_condition = new_condition

                            else:
                                replacement = rand_var[cond_idx]
                                new_condition = self.replace_condition(old_condition, replacement, random_index)
                            new_plan.append(new_condition)

                new_plans.append(new_plan)
        return new_plans
    




  


# External libraries
from z3 import *                # Requires: pip install z3-solver
import numpy as np

# Core system modules (planet package)
from planet.candl import *          #
from planet.helpers import *       
from planet.narray import *
from planet.solver import BitVecSolver, TestSolver
from planet.constraint import (
    StartWith, Counterbalance, NoRepeat,
    InnerBlock, OuterBlock, Constraint,
    SetRank, SetPosition, AbsoluteRank, Order
)
from planet.variable import MultiFactVariable
from planet.unit import Units

# NOTE: combine this with des pls
class Designer:
    """The assignment class matches every unit to an order of conditions
        based on constraints set on unit variables, such as block factors
        of participants, pid, and the state of observation of a subject. 

    """
    def __init__(self):
        self.constraints = []
        self.previous_snapshot = None
        self.previous_model = []

    def start(self, design):
        self.design = design
        self.num_plans = design.num_plans()
        self.num_trials = design.get_width()
        self.variables = design.variables
        self.shape = self.determine_shape()
        self.solver = BitVecSolver(self.shape, self.variables)
        self.check_trial_compatibility()
    
    
    def check_trial_compatibility(self):
        if self.design.maximum_trials > -1 and self.num_trials > self.design.maximum_trials:
            raise ValueError(
                f"Design specifies {self.num_trials} trial(s), "
                f"but there are only {self.design.maximum_trials} possible conditions. "
                "Reduce trials or use nesting to repeat conditions."
            )
        
        if self.design._minimum_trials > 1 and self.num_trials < self.design._minimum_trials:
            raise ValueError(
                f"Design specifies {self.num_trials} trial(s), "
                f"but there must be {self.design._minimum_trials} trial(s). "
                "This is often due to nesting or crossing designs."
            )

    @property
    def design_has_changed(self):
        return self.design.snapshot() != self.previous_snapshot


    def _get_experiment_variables(self, constraint):
        if isinstance(constraint.variable, MultiFactVariable):
            return constraint.get_variable().get_variables()
        else:    
            return [constraint.get_variable()]

    def eval_constraints(self, constraints, groups, width):
        # Process constraints
        for constraint in constraints:
            assert isinstance(constraint, Constraint)
            variable = self._get_experiment_variables(constraint)

            match constraint:
                
                case Counterbalance():
                
                    if groups:
                        constraint.width = (
                            constraint.width if constraint.width else width
                        )
                        constraint.height = (
                            constraint.height if constraint.height else groups
                        )

                    self.counterbalance(
                        variable,
                        w=constraint.width,
                        h=constraint.height,
                        stride=constraint.stride,
                    )


                case NoRepeat():
                        self.solver.all_different(
                            variable, 
                            constraint.width, 
                            constraint.stride
                        )

                case StartWith():
                    self.start_with(
                        constraint.variable, 
                        constraint.condition
                    )
                
                case SetPosition():
                    self.set_position(
                        constraint.variable, 
                        constraint.condition,
                        constraint.position
                    )

                case SetRank():
                    constraint.width = (constraint.width if constraint.width else width)
                    self.set_rank(
                        constraint.variable, 
                        constraint.condition,
                        constraint.rank,
                        constraint.condition2,
                        constraint.width,
                        constraint.stride
                    )

                case AbsoluteRank():
                    self.absolute_rank(
                        constraint.variable, 
                        constraint.ranks,
                        constraint.width,
                        constraint.stride
                    )

                case Order():
                    self.order(
                        constraint.variable, 
                        constraint.sequence,
                        constraint.width,
                        constraint.stride
                    )

                case InnerBlock():
                    constraint.width = (
                        constraint.width if constraint.width else width
                    )
                    self.match_inner(
                        constraint.variable, 
                        constraint.width, 
                        constraint.height
                    )

                case OuterBlock():
                    self.match_outer(
                        constraint.variable, 
                        constraint.width, 
                        constraint.height
                    )
 
    
    def determine_shape(self):
        if self.num_plans: 
            n = self.num_plans
        else: 
            n = 1
        return tuple([n, self.num_trials])
    
    # FIXME: creating block matrix for specific test case 
    # Note: use for creating blocks

    # NEED TO DECOUPLE THIS
    def match_inner(self, variable, width, height):
        # get number of block matrices per column
        n = int(self.shape[0] / height)
        # get number of block matrices per row
        m = int(self.shape[1] / width)

    
        for i in range(n):
            for j in range(m):
                self.solver.match_block(
                    variable, 
                    [
                        (i*height + 0, i * height  + height, 1)
                        , (j*width + 0, j * width + width, 1)
                    ]
                )


    def match_outer(self, v, w, h):
        for i in range(h):
            for j in range(w):
                
                self.solver.match_block(v, [(i, self.shape[0], h), (j, self.shape[1], w)])

    def counterbalance(self, v, w, h, stride = [1, 1]):
        block = [(0, h, stride[0]), (0, w, stride[1])]
        self.solver.counterbalance(block, v)

    def start_with(self, variable, condition):
        self.solver.start_with(variable, variable.conditions.index(condition))

    def set_rank(self, variable, condition, rank, condition2):
        self.solver.set_rank(variable, variable.conditions.index(condition), rank, variable.conditions.index(condition2), width, stride)
    
    def set_position(self, variable, condition, pos):
        self.solver.set_position(variable, variable.conditions.index(condition), pos)
    
    def absolute_rank(self, variable, ranks, width, stride):
        transformed_ranks = {variable.conditions.index(condition): rank for condition, rank in ranks.items()}
        self.solver.absolute_rank(variable, transformed_ranks, width, stride)

    def order(self, variable, sequence, width, stride):
        transformed_sequence = [variable.conditions.index(condition) for condition in sequence]

        self.solver.order(variable, transformed_sequence, width, stride)

    # NOTE: this is with a bitvec representation...
    # ensure that this works
    def eval(self):
        if not self.design_has_changed:
            model = self.previous_model

        else:
            self.eval_constraints(self.design.get_constraints(), self.num_plans, self.num_trials)
            # perhaps this is where we create a different class?
            # we have so many new instance vars
            # these should maybe be classes or something? 
   
            model = self.solver.get_one_model()
            self.previous_snapshot = self.design.snapshot()
            self.previous_model = model
        return self.decode(model)
  
    def decode(self, model):
        if not len(model):
            return np.array([])
        else:
            reshaped_model = np.array(model).reshape(self.shape).tolist()
            return np.array(self.solver.encoding_to_name(reshaped_model, self.variables))
        
  
    # NOTE: this is with a bitvec representation...
    # ensure that this works
    def eval_all(self):
        # perhaps this is where we create a different class?
        # we have so many new instance vars
        # these should maybe be classes or something? 
        if self.num_plans:
            models = self.solver.get_all_models()

            ret = []
            for model in models:
                assert len(model) > 0
                model = np.array(model).reshape(self.shape).tolist()
                ret.append(np.array(self.solver.encoding_to_name(model, self.variables)))

            return ret
        else:
            model = self.solver.get_all_models()
            self.get_groups(model)
            return np.array(self.solver.encoding_to_name(model, self.variables))
            

    


from planet.constraint import AbsoluteRank, Counterbalance, Order
from typing import Callable

class ConstraintManager:
    def __init__(self):
        self.constraints = []  # global list of all constraint instances
        self.rank_constraints = {}


    def add_absolute_rank(self, variable, condition, rank):
        if variable not in self.rank_constraints:
            constraint = AbsoluteRank(variable, condition, rank)
            self.rank_constraints[variable] = constraint
            self.constraints.append(constraint)
        else:
            constraint = self.rank_constraints[variable]
            constraint.add_rank(condition, rank)

        return constraint
    


    def add_constraint(self, constraint):
        # Avoid duplicates
        if self.has_constraint(constraint.variable, type(constraint)):
            return

        self.constraints.append(constraint)


    def get_constraints_for_variable(self, variable):
        return [
            c for c in self.constraints if c.variable == variable
        ]

    def has_constraint(self, variable, constraint_type):
        return any(
            isinstance(c, constraint_type) and c.variable == variable
            for c in self.constraints
        )
    
    def check_property(self, property:Callable):
        return any(property(c) for c in self.constraints)
    
    def stringified(self):
        return "\n".join(str(c) for c in self.constraints)
    
    def get_constraints(self):
        return self.constraints
    
    def get_counterbalanced_constraints(self):
        """
        Get all counterbalanced constraints.

        Returns:
            list: A list of all Counterbalance constraints.
        """
        return [c for c in self.constraints if isinstance(c, Counterbalance)]



from dataclasses import dataclass

@dataclass
class DesignRegion:
    """Represents a region in the design space. For example, 
    it might represent a subset of the first 2 trials in a 4 trial design."""
    width:int
    height:int
    stride:list[int]


from z3 import *
import numpy as np
from .candl import *
from .helpers import *
from .bitvector import BitVectors, TupleVariables
import itertools
import warnings
from time import time
set_param('parallel.enable', True)

class Solver: 
    def __init__(self, shape, variables):
        self.shape = shape
        self.variables = variables
        self.solver = z3.Optimize()

class BitVecSolver(Solver):
    def __init__(self, shape, variables):
        """
        Solver using Z3 BitVectors .
        Automatically sets up BitVectors and enforces row uniqueness.
        """
        super().__init__(shape, variables)
        self.bitvectors = BitVectors(np.prod(self.shape), self.variables)
        self.z3_variables = self.bitvectors.get_variables()
        self.constrain_z3_values()
        self.distinguish_rows()

    # you can come up with a better name
    def constrain_z3_values(self):
        """
        Enforce that each Z3 variable can only take a valid level
        based on its corresponding experimental variable.
        """
        for variable in self.variables:
            for index in range(np.prod(self.shape)):
                lo = 0
                hi = len(variable) - 1
                masked_bits = self.bitvectors.get_variable_assignment(variable, self.z3_variables[index])
                self.solver.add(
                        And(
                            lo <= masked_bits, 
                            hi >= masked_bits
                        )  
                )

    # you can come up with a better name
    def enforce_ordering(self):
        """
        Enforce that each Z3 variable can only take a valid level
        based on its corresponding experimental variable.
        """
        for variable in self.variables:
            for index in range(0, np.prod(self.shape)-self.shape[1], self.shape[1]*2):
                masked_bits1 = self.bitvectors.get_variable_assignment(variable, self.z3_variables[index])
                masked_bits2 = self.bitvectors.get_variable_assignment(variable, self.z3_variables[index+self.shape[1]])

                self.solver.add(
                        masked_bits1 <= masked_bits2
                )


    # NOTE: no rows can repeat in a given matrix 
    def _enforce_ordering(self):
        plans = shape_array(self.z3_variables, self.shape)
        plan_encodings = []

        for plan in plans:
            # Concatenate all variables in the row into one BitVec
            plan_bits = plan[0]
            for var in plan[1:]:
                plan_bits = Concat(plan_bits, var)
            plan_encodings.append(plan_bits)

        for i in range(len(plan_encodings)-1):
            self.solver.add(
                plan_encodings[i] < plan_encodings[i+1]
            )
    
  
    def get_partition(self, width, stride):
        """
        Partition the design matrix by columns if width is specified.

        Args:
            width (int): Number of columns per partition.
            stride (int): Step size between partitions.

        Returns:
            np.ndarray: Partitioned design matrix.
        """
        dim_variables = shape_array(self.z3_variables, self.shape)
        if width is not None:
            dim_variables = partition_matrix_by_columns(dim_variables, width, stride)
        return dim_variables

    def all_different(self, variables=None, width = None, stride = 1):
        """
        Enforce that for each row in the partitioned matrix, 
        all combinations of the selected variables differ in at least one position.
        """
        matrix = self.get_partition(width, stride)
        for row in matrix:
            row_assignments = [
                [self.bitvectors.get_variable_assignment(v, r) for v in variables] 
                for r in row
            ]
            self.solver.add(at_least_one_diff(row_assignments))

    def _all_different_counterbalance(self, variables=None, block=[]):
        """
        Enforce that for each row in the partitioned matrix, 
        all combinations of the selected variables differ in at least one position.
        """
        matrix = shape_array(self.z3_variables, self.shape)
        matrix = self.slice_matrix(matrix, block)
 
        for i in range(0, len(matrix), len(variables[0])):
            block = matrix[i:i+len(variables[0])].transpose()
            for row in block:
                row_assignments = [
                    [self.bitvectors.get_variable_assignment(v, r) for v in variables] 
                    for r in row
                ]
                self.solver.add(at_least_one_diff(row_assignments))
        
    def n_trials(self):
        return self.shape[1]

    # NOTE: no rows can repeat in a given matrix 
    def distinguish_rows(self):
        plans = shape_array(self.z3_variables, self.shape)
        plan_encodings = []

        for plan in plans:
            # Concatenate all variables in the row into one BitVec
            plan_bits = plan[0]
            for var in plan[1:]:
                plan_bits = Concat(plan_bits, var)
            plan_encodings.append(plan_bits)

        # Enforce that all rows are distinct
        self.solver.add(Distinct(*plan_encodings))


    def slice_matrix(self, arr, block = []):
        # Unpack block and slice matrix
        (start_col, end_col, col_step), (start_row, end_row, row_step) = block
        return np.array(arr)[start_col:end_col:col_step, start_row:end_row:row_step]


    def absolute_rank(self, var, ranks, width, stride):
        design_matrix = self.get_partition(width, stride)
        n = len(var)

        rank_fn = Function(f'rank_{str(var)}', BitVecSort(self.bitvectors.determine_num_bits()), IntSort())
        self.solver.add([rank_fn(plan[i]) >= rank_fn(plan[i+1]) for i in range(n-1) for plan in design_matrix])

        # for use in quantified expression
        x = self.bitvectors.new_bitvector()
        for condition, rank in ranks.items():
            # set the rank for every condition of the respective variable
            self.solver.add(
                ForAll([x],
                    Implies(
                        self.bitvectors.get_variable_assignment(var, x) == condition,
                        rank_fn(x) == rank
                    )
                )
            )
    
    def order(self, var, sequence, width, stride):
        """
        Enforce that every row of the design matrix has the same sequence of conditions
        for the specified variable.

        Args:
            var: The variable to enforce the order on.
            sequence: A list specifying the sequence of conditions.
            width: The number of columns per partition.
            stride: The step size between partitions.
        """
        # Partition the design matrix
        design_matrix = self.get_partition(width, stride)

        # Enforce the sequence of conditions for each row
        for row in design_matrix:
            for i, condition in enumerate(sequence):
                self.solver.add(
                    self.bitvectors.get_variable_assignment(var, row[i]) == condition
                )
            
    def count(self, variables, condition, f):
        """
        if f returns true increase count by one for all variables
        """
        counts = [If(f(var, condition), BitVecVal(1, 2), BitVecVal(0, 2)) for var in variables]
        return sum(counts)
    
    
    def precompute_complete_counterbalancing(self, conditions, plans, variables):
        '''THis is an optimization because there is only one canonical solution
        for fully-counterbalanced designs. There are many symmetrical solutions
        that z3 needs to explore. Using this, we precompute the single
        solution.'''
  
        is_complete = lambda conds, rows: len(rows) == math.factorial(len(conds))
        if is_complete(conditions, plans):
            permutations = list(itertools.permutations(conditions))
            for i in range(len(plans)):
                for j in range(len(plans[i])):
                    # Need to set conditions of all counterbalanced variables
                    for k in range(len(variables)):
                        self.solver.add(self.bitvectors.get_variable_assignment(variables[k], plans[i][j]) == permutations[i][j][k])
    

    def counterbalance(self, block=[], variables=None):
        possible_conditions = list(itertools.product(
            *[set(range(len(variable))) for variable in variables]
        ))
        design_matrix = shape_array(self.z3_variables, self.shape)
        design_matrix = self.slice_matrix(design_matrix, block)
        design_matrix_transposed = np.transpose(design_matrix)

        counts = []
        for trials in design_matrix_transposed:
            masked_variables = list(zip(*[
                list(map(lambda z3_variable: self.bitvectors.get_variable_assignment(variable, z3_variable), trials)) 
                for variable in variables
            ]))
            
            check_equality = lambda variables, assignments: And([variables[i] == assignments[i] for i in range(len(variables))])
            counts.extend([self.count(masked_variables, condition, check_equality) for condition in possible_conditions])
        
        # # Add constraints to ensure equal counts for all combinations
        for i in range(len(counts)):
            self.solver.add(counts[i] == counts[0])

        self.precompute_complete_counterbalancing(possible_conditions, design_matrix, variables) 
    

   
    def match_block(self, variable, block = []):
        plans = shape_array(self.z3_variables, self.shape)
        block = self.slice_matrix(plans, block)

        flat_block = np.array(block).flatten()
        for i in range(len(flat_block)):
            for j in range(i, len(flat_block)):
                a1 = self.bitvectors.get_variable_assignment(variable, flat_block[i])
                a2 = self.bitvectors.get_variable_assignment(variable, flat_block[j])
                self.solver.add(a1==a2)

        # works when z3 representation is a matrix 
    def get_one_model(self):
        all_assignments = []

        if (self.solver.check() == sat):
            all_assignments = []
            model = self.solver.model()
      
            for var in self.z3_variables:
                all_assignments.append(model.evaluate(model[var]))
 
        return all_assignments
    
    # works when z3 rep iterates through many arrays
    # NOTE: should I merge these two functions?
    def get_all_models(self):
        all_orders = []
        self.solver.push()
        count = 0
        capacity = 10000
        while self.solver.check() == sat and count < capacity:
            model = self.solver.model()
            block = []
            order = []
            for var in self.z3_variables:
                block.append(var != model[var])
                order.append(model.evaluate(model[var]))

            all_orders.append(order)
            self.solver.add(Or(block))
            count+=1
        if self.solver.check() == sat:
            warnings.warn(f"There are more orders not captured in the model due order capacity (currently set to {capacity})")

        self.solver.pop()
        return all_orders


    def encoding_to_name(self, all_orders, variables):
        # this makes life much easier
        z3_assignments = flatten_array(all_orders)
        num_vars = len(variables)
        
        orders = []
        for assignment in z3_assignments:
            # this is an integer value, represented as a binary string
            assignment = int(str(assignment))
            # this is where we will store the string rep of a condition
            # which is a concatination of it's variable assignments
            decoded_assignment = ""
 
            # # there is probably a more efficient way to do this
            for i in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment 
                # shift right variable assignment unitl desired var assignment 
                # is in the lower bits (+1 to account for two's compliment)
         
                for j in range(i):
                    variable_assignment >>= get_num_bits(len(variables[j])) + 1

                # get lower bits (associated with given variable)
                mask =  all_ones(get_num_bits(len(variables[i]))+ 1)
                variable_assignment &= mask

                # FIXME
                if variable_assignment >= len(variables[i]):
                    condition = "dropout"
                else:
              
                    # FIXME: this is silly, but need to inverse order to match array indexing
                    # variable_assignment = len(variables[i]) - 1 - variable_assignment
                    condition = variables[i].conditions[variable_assignment]

                # conder variable assignment to a string representation 
                decoded_assignment+=str(condition)
                # FIXME; hyphenate output for readability
                if i < len(variables)-1:
                    decoded_assignment+="-"

            orders.append(decoded_assignment)

        # return back in original form
        return shape_array(orders, np.array(all_orders).shape)
    
    def name_to_encoding(self, model):

        # this makes life much easier
        model_assignments = flatten_array(model)
        num_vars = len(self.variables)
        
        for i in range(len(model_assignments)):
            assignment = model_assignments[i].split("-")

            # there is probably a more efficient way to do this
            for j in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment[j]
                condition = self.variables[j].condition_map[variable_assignment]
                __z3__ = self.bitvectors.get_variable_assignment(self.variables[j], self.z3_variables[i])
                self.solver.add(__z3__ == condition)

    def check_model(self, model):
  
        self.solver.push()

        self.name_to_encoding(model, self.variables)
        result = self.solver.check()

        self.solver.pop()

        return result



class TestSolver(Solver):
    def __init__(self, shape, variables):
        """
        Solver using Z3 BitVectors .
        Automatically sets up BitVectors and enforces row uniqueness.
        """
        super().__init__(shape, variables)
        self.tuples = TupleVariables(np.prod(self.shape), self.variables)
        self.z3_variables = self.tuples.get_variables()
        self.constrain_z3_values()
        self.distinguish_rows()

    # you can come up with a better name
    def constrain_z3_values(self):
        """
        Enforce that each Z3 variable can only take a valid level
        based on its corresponding experimental variable.
        """
        for variable in self.variables:
            for index in range(np.prod(self.shape)):
                lo = 0
                hi = len(variable) - 1
                variable_assignment = self.tuples.get_variable_assignment(variable, self.z3_variables[index])
                self.solver.add(
                        And(
                            lo <= variable_assignment, 
                            hi >= variable_assignment
                        )  
                )

    # you can come up with a better name
    def enforce_ordering(self):
        """
        Enforce that each Z3 variable assignment must be within range of 
        possible conditions for the respective variable. 
        """
        for variable in self.variables:
            # make this loop more readable 
            for index in range(0, np.prod(self.shape)-self.shape[1], self.shape[1]*2):
                variable_assignment1 = self.tuples.get_variable_assignment(variable, self.z3_variables[index])
                variable_assignment2 = self.tuples.get_variable_assignment(variable, self.z3_variables[index+self.shape[1]])

                self.solver.add(
                        variable_assignment1 <= variable_assignment2
                )

    def get_partition(self, width, stride):
        """
        Partition the design matrix by columns if width is specified.

        Args:
            width (int): Number of columns per partition.
            stride (int): Step size between partitions.

        Returns:
            np.ndarray: Partitioned design matrix.
        """
        dim_variables = shape_array(self.z3_variables, self.shape)
        if width is not None:
            dim_variables = partition_matrix_by_columns(dim_variables, width, stride)
        return dim_variables

    def all_different(self, variables=None, width = None, stride = 1):
        """
        Enforce that for each row in the partitioned matrix, 
        all combinations of the selected variables differ in at least one position.
        """
        matrix = self.get_partition(width, stride)
        for row in matrix:
            row_assignments = [
                [self.tuples.get_variable_assignment(v, r) for v in variables] 
                for r in row
            ]
            self.solver.add(at_least_one_diff(row_assignments))

    def _all_different_counterbalance(self, variables=None, block=[]):
        """
        Enforce that for each row in the partitioned matrix, 
        all combinations of the selected variables differ in at least one position.
        """
        matrix = shape_array(self.z3_variables, self.shape)
        matrix = self.slice_matrix(matrix, block)
 
        for i in range(0, len(matrix), len(variables[0])):
            block = matrix[i:i+len(variables[0])].transpose()
            for row in block:
                row_assignments = [
                    [self.bitvectors.get_variable_assignment(v, r) for v in variables] 
                    for r in row
                ]
                self.solver.add(at_least_one_diff(row_assignments))
        
    def n_trials(self):
        return self.shape[1]

    # NOTE: no rows can repeat in a given matrix 
    def distinguish_rows(self):
        plans = shape_array(self.z3_variables, self.shape)

        n = len(plans)
        m = len(plans[0])
        constraints = []
        for i in range(n):
            for j in range(i + 1, n):
                constraints.append(
                    Or([plans[i][k] != plans[j][k] for k in range(m)])
        )
        self.solver.add(And(constraints))


    def slice_matrix(self, arr, block = []):
        # Unpack block and slice matrix
        (start_col, end_col, col_step), (start_row, end_row, row_step) = block
        return np.array(arr)[start_col:end_col:col_step, start_row:end_row:row_step]


    def absolute_rank(self, var, ranks, width, stride):
        design_matrix = self.get_partition(width, stride)
        n = len(var)

        rank_fn = Function(f'rank_{str(var)}', BitVecSort(self.bitvectors.determine_num_bits()), IntSort())
        self.solver.add([rank_fn(plan[i]) >= rank_fn(plan[i+1]) for i in range(n-1) for plan in design_matrix])

        # for use in quantified expression
        x = self.bitvectors.new_bitvector()
        for condition, rank in ranks.items():
            # set the rank for every condition of the respective variable
            self.solver.add(
                ForAll([x],
                    Implies(
                        self.bitvectors.get_variable_assignment(var, x) == condition,
                        rank_fn(x) == rank
                    )
                )
            )
            
    def count(self, variables, condition, f):
        """
        if f returns true increase count by one for all variables
        """
        counts = [If(f(var, condition), 1, 0) for var in variables]
        return sum(counts)
    
    
    def precompute_complete_counterbalancing(self, conditions, plans, variables):
        '''THis is an optimization because there is only one canonical solution
        for fully-counterbalanced designs. There are many symmetrical solutions
        that z3 needs to explore. Using this, we precompute the single
        solution.'''
  
        is_complete = lambda conds, rows: len(rows) == math.factorial(len(conds))
        if is_complete(conditions, plans):
            permutations = list(itertools.permutations(conditions))
            for i in range(len(plans)):
                for j in range(len(plans[i])):
                    # Need to set conditions of all counterbalanced variables
                    for k in range(len(variables)):
                        self.solver.add(self.tuples.get_variable_assignment(variables[k], plans[i][j]) == permutations[i][j][k])
    

    def counterbalance(self, block=[], variables=None):
        possible_conditions = list(itertools.product(
            *[set(range(len(variable))) for variable in variables]
        ))
        design_matrix = shape_array(self.z3_variables, self.shape)
        design_matrix = self.slice_matrix(design_matrix, block)
        design_matrix_transposed = np.transpose(design_matrix)

        counts = []
        for trials in design_matrix_transposed:
            masked_variables = list(zip(*[
                list(map(lambda z3_variable: self.tuples.get_variable_assignment(variable, z3_variable), trials)) 
                for variable in variables
            ]))
            
            check_equality = lambda variables, assignments: And([variables[i] == assignments[i] for i in range(len(variables))])
            counts.extend([self.count(masked_variables, condition, check_equality) for condition in possible_conditions])
        
        # # Add constraints to ensure equal counts for all combinations
     
        for i in range(len(counts)):
            self.solver.add(counts[i] == counts[0])

        self.precompute_complete_counterbalancing(possible_conditions, design_matrix, variables) 
    

   
    def match_block(self, variable, block = []):
        plans = shape_array(self.z3_variables, self.shape)
        block = self.slice_matrix(plans, block)

        flat_block = np.array(block).flatten()
        for i in range(len(flat_block)):
            for j in range(i, len(flat_block)):
                a1 = self.tuples.get_variable_assignment(variable, flat_block[i])
                a2 = self.tuples.get_variable_assignment(variable, flat_block[j])
                self.solver.add(a1==a2)

    # works when z3 representation is a matrix 
    def get_one_model(self):
        all_assignments = []
        t1 = time()
        
        if self.solver.check() == sat:
            model = self.solver.model()
            
            for z3_tuple in self.z3_variables:
                # evaluate each field in the tuple
                evaluated_fields = [
                    model.eval(proj(z3_tuple), model_completion=True).as_long()
                    for proj in self.tuples.tuple_proj
                ]
                all_assignments.append(tuple(evaluated_fields))
        
        t2 = time()
        print(f"Time taken to get one model: {t2 - t1} seconds")
        print(all_assignments)
        return all_assignments


    def encoding_to_name(self, all_orders, variables):
        # this makes life much easier
        z3_assignments = flatten_array(all_orders)
        num_vars = len(variables)
        
        orders = []
        for assignment in z3_assignments:
            # this is an integer value, represented as a binary string
            # this is where we will store the string rep of a condition
            # which is a concatination of it's variable assignments
            decoded_assignment = ""
 
            # # there is probably a more efficient way to do this
            for i in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment 
                # shift right variable assignment unitl desired var assignment 
                # is in the lower bits (+1 to account for two's compliment)
         
                for j in range(i):
                    variable_assignment >>= get_num_bits(len(variables[j])) + 1

                # get lower bits (associated with given variable)
                mask =  all_ones(get_num_bits(len(variables[i]))+ 1)
                variable_assignment &= mask

                # FIXME
                if variable_assignment >= len(variables[i]):
                    condition = "dropout"
                else:
              
                    # FIXME: this is silly, but need to inverse order to match array indexing
                    # variable_assignment = len(variables[i]) - 1 - variable_assignment
                    condition = variables[i].conditions[variable_assignment]

                # conder variable assignment to a string representation 
                decoded_assignment+=str(condition)
                # FIXME; hyphenate output for readability
                if i < len(variables)-1:
                    decoded_assignment+="-"

            orders.append(decoded_assignment)

        # return back in original form
        return shape_array(orders, np.array(all_orders).shape)
    
    def name_to_encoding(self, model):

        # this makes life much easier
        model_assignments = flatten_array(model)
        num_vars = len(self.variables)
        
        for i in range(len(model_assignments)):
            assignment = model_assignments[i].split("-")

            # there is probably a more efficient way to do this
            for j in range(num_vars):
                # get the ith variable's assignment
                variable_assignment = assignment[j]
                condition = self.variables[j].condition_map[variable_assignment]
                __z3__ = self.variables.get_variable_assignment(self.variables[j], self.z3_variables[i])
                self.solver.add(__z3__ == condition)

    def check_model(self, model):
  
        self.solver.push()

        self.name_to_encoding(model, self.variables)
        result = self.solver.check()

        self.solver.pop()

        return result


from planet.design import Design
from itertools import combinations
from planet.variable import ExperimentVariable, MultiFactVariable
import warnings

class Analysis:
    """Main class for analyzing experimental designs."""
    def __init__(self, design:Design):

        if design.variables is None:
            raise AttributeError("This design has no variables!")

        self.design = design
        self.main_effects = set()
        self.interaction_effects = set()
        self.time_varying_effects = set()
        self.ws_comparisons = set()

        self.counterbalanced_variables = [
            design_variable
            for design_variable in self.design.design_variables.values()
            if design_variable.is_counterbalanced or design_variable.is_random
        ]

        self.perform_analysis()


    def analyze_main_effects(self):
        """Update main effects based on the design's variables. The main effect
        is estimable if it is a variable in the design and there are more
        than two conditions for that variable. This is under the assumption that
        all plans are distinct and each plan is assigned an equal number of times.
        """
    
        for var, dvar in self.design.design_variables.items():
            if dvar.is_counterbalanced or dvar.is_random: 
                if len(var.conditions) >= 2:
                    self.main_effects.update(var.get_variables())
            elif not dvar.is_repeated and dvar.is_ordered:
                if len(var.conditions) >= 2:
                    self.main_effects.update(var.get_variables())
            else:
                warnings.warn(f"Could not perform analysis for variable {var}.")
    

    def analyze_interaction_effects(self):  
        """Analyze and update interaction effects based on the design's variables.

        An interaction effect is estimable if:
        - all of its component main effects are estimable
        - the number of plans is divisible by the product of the number of
        conditions for each variable in the interaction (i.e., the interaction
        conditions are assigned an equal number of times)
        - the combined main effects appear an equal number of times across all
        assignments. 
        """

        # First, we want to identify all multivariable combinations that are
        # explicitly counterbalanced in the design. 
        # NOTE: need to check properties with rankings.
        # NOTE: we only want pairwise combinations, and we can infer this from
        # interactiosn with three or more variables. This is a TODO. 
        for var, dvar in self.design.design_variables.items():
            
            if dvar.is_counterbalanced or dvar.is_random:
                if len(var.conditions) >= 2:
                    if var.is_multifact():
                        # Add new multifact variables for all pairs of constituent variables
                        constituent_vars = var.get_variables()  # or however you access the component variables
                        for v1, v2 in combinations(constituent_vars, 2):
                            pair_var = MultiFactVariable([v1, v2])  # replace with your actual constructor
                            self.interaction_effects.add(pair_var)
                    else: 
                        self.interaction_effects.add(var)
            elif not dvar.is_repeated and dvar.is_ordered:
                if len(var.conditions) >= 2:
                    if var.is_multifact():
                        # Add new multifact variables for all pairs of constituent variables
                        constituent_vars = var.get_variables()  # or however you access the component variables
                        for v1, v2 in combinations(constituent_vars, 2):
                            pair_var = MultiFactVariable([v1, v2])  # replace with your actual constructor
                            self.interaction_effects.add(pair_var)
                    else: 
                        self.interaction_effects.add(var)
            else:
                warnings.warn(f"Could not perform analysis for variable {var}.")

       # Then, we want to identify all implicit combinations created by adding
       # individual variables to a design. This occurs when two variables are
       # added independently to a design, and the plan limit is less than the
       # maximum number of plans. 

        # What is the effect of including precomputed plans? I think just
        # redundant computation.
        plan_count, _ = self.design._calculate_maximum_plans()
        plan_limit = self.design.num_groups

            # Add every pair of variables as a multifact variable
        for var1, var2 in combinations(self.design.design_variables, 2):
            var1_dv = self.design.design_variables[var1]
            var2_dv = self.design.design_variables[var2]
            # at least on variable must be unranked 
            if (var1_dv.is_counterbalanced or var1_dv.is_random) and (var2_dv.is_counterbalanced or var2_dv.is_random):
                if (plan_count <= plan_limit or plan_limit == 0):
                    multifact_var = MultiFactVariable([var1, var2])
                    self.interaction_effects.add(multifact_var)

            elif ((var1_dv.is_counterbalanced or var1_dv.is_random) and var2_dv.is_ordered) or ((var2_dv.is_counterbalanced or var2_dv.is_random) and var1_dv.is_ordered):
                multifact_var = MultiFactVariable([var1, var2])
                self.interaction_effects.add(multifact_var)
            
        # Lastly, we want to identify all implicit combinations created by
        # composing designs. 
        # - get all counterbalanced variables with outer block. 
        # - get all variables with inner block. 
        # - For each pair, check if inner block is within the bounds of outer block.
        inner_variables = [
            (design_var, design_var.constraint_spec["InnerBlock"])
            for design_var in self.design.design_variables.values()
            if design_var.is_blocked_inner
        ]

        outer_variables = [
            (design_var, design_var.constraint_spec["OuterBlock"])
            for design_var in self.design.design_variables.values()
            if design_var.is_blocked_outer
        ]
 
        # NOTE: this checks for nest! Still need to check for cross ;) 
        for outer_var, outer_spec in outer_variables:
            for inner_var, inner_spec in inner_variables:
                # the inner block height is always a multiple or factor of the
                # outer block height based on how we compose designs. We always
                # add an inner block when we add an outer block. 
                if (
                    (outer_spec.height <= inner_spec.height 
                    and outer_spec.variable != inner_spec.variable) 
                ):

                    outer_repeats = outer_var.constraint_spec["NoRepeat"]
                    inner_repeats = outer_var.constraint_spec["NoRepeat"]
                

                    outer_check = (
                        ((outer_repeats is not None and 
                        outer_repeats.width*outer_repeats.stride < outer_spec.width and outer_var.is_ordered) and self.design.num_trials > outer_spec.width*outer_spec*outer_spec.stride)
                    ) 

                    inner_check = (
                          ((inner_repeats is not None and 
                        inner_repeats.width*inner_repeats.stride < inner_spec.width and inner_var.is_ordered) and self.design.num_trials > inner_spec.width*inner_spec*inner_spec.stride)
                    ) 

                    if (outer_spec.width <= inner_spec.width and inner_spec.height <= outer_spec.height) and (outer_check or (outer_var.is_counterbalanced or outer_var.is_random)) and (inner_check or (inner_var.is_counterbalanced or inner_var.is_random)): 
                        interaction = MultiFactVariable([outer_var.variable, inner_var.variable])
                        self.interaction_effects.add(interaction)

                    elif (inner_spec.height <= outer_spec.height) and (outer_check or (outer_var.is_counterbalanced or outer_var.is_random)) and (inner_var.is_counterbalanced or inner_var.is_random):
                        interaction = MultiFactVariable([outer_var.variable, inner_var.variable])
                        self.interaction_effects.add(interaction)
                        

            

        # For cross, need to check that counterbalance is on same width as the
        # inner block, and that the either the number of trials is the same as
        # the number of conditions or there is an outer block with a width the
        # same as the number of conditions! 
       
    def analyze_time_varying_effects(self):
        """Analyze and update time-varying effects based on the design's variables.

        A time-varying effect is estimable if:
        - the primary effect is estimable
        - The variable is counterbalanced (i.e., its conditions are evenly distributed across trials).
        - If the variable is a multifact variable, its sub-variables are also
        time-varying.
        
        Joint-variables (i.e., multifact variables) may not be explicitly
        counterbalanced. In these cases, we can infer counterbalancing if the
        individual variables are counterbalanced and they have overlapping inner
        and outer blocks, respectively. 
        """

        # Analyze time-varying effects
        if self.design.get_width() > 1:
            for counterbalanced_variable in self.counterbalanced_variables:
                if counterbalanced_variable.get_variable().is_multifact(): 
                    self.time_varying_effects.update(counterbalanced_variable.get_variable().get_variables())
                    subvars = counterbalanced_variable.get_variable().get_variables()
                    self.time_varying_effects.update(subvars)
                    for var_a, var_b in combinations(subvars, 2):
                        self.time_varying_effects.add(MultiFactVariable([var_a, var_b]))
                else: 
                    self.time_varying_effects.add(counterbalanced_variable.get_variable())
                

        for var1, var2 in combinations(self.counterbalanced_variables, 2):
            if (self.design.get_width() > 1): 
                combined_var = MultiFactVariable([var1.variable, var2.variable])
                if combined_var in self.interaction_effects:
                    self.time_varying_effects.add(combined_var)


    def analyze_ws_comparisons(self):
        """
        Determine whether the variable is compared within-subjects. This is true
        if every participant is exposed to *every* condition of the variable. A
        variable can be added as within-subjects and lack comprehensive
        comparisons if the number of trials is less than the number of
        conditions. 
        """

        for var in self.design.design_variables.values():
            
            no_repeat = var.constraint_spec.get("NoRepeat")
            if no_repeat is None:
                continue  # Skip variables without a NoRepeat constraint

            width = no_repeat.width
            stride = no_repeat.stride

            if (width/stride) % len(var.variable) == 0 and self.design.get_width() % len(var.variable) == 0:
                subvars = var.get_variable().get_variables()
                self.ws_comparisons.update(subvars)
                for var_a, var_b in combinations(subvars, 2):
                    self.ws_comparisons.add(MultiFactVariable([var_a, var_b]))
                self.ws_comparisons.update(var.get_variable().get_variables())

        inner_variables = [
            design_var.constraint_spec["InnerBlock"]
            for design_var in self.design.design_variables.values()
            if design_var.is_blocked_inner and not design_var.is_repeated
        ]

        outer_variables = [
            design_var.constraint_spec["OuterBlock"]
            for design_var in self.design.design_variables.values()
            if design_var.is_blocked_outer and not design_var.is_repeated
        ]

        for outer_var in outer_variables:
            for inner_var in inner_variables:
                if outer_var.width <= inner_var.width and self.design.trials % outer_var.width * inner_var.width == 0 and outer_var != inner_var:
             
                    interaction = MultiFactVariable([outer_var.variable, inner_var.variable])
                    self.ws_comparisons.add(interaction)
                    self.ws_comparisons.update(interaction.get_variables())
                

    def perform_analysis(self):

        if self.design.is_random:
            str = "There are random variables in this design, which means that the analysis is subject to random variation and is more likely to hold in larger samples. Consider adding counterbalancing constraints to ensure that the analysis holds in smaller samples.\nVariables that are completely randomized:\n"
           

            for var in self.design.variables: 
                str += f"\t{var}\n"

            warnings.warn(str)


        self.analyze_main_effects()
        self.analyze_interaction_effects()
        self.analyze_time_varying_effects()
        self.analyze_ws_comparisons()

    def __str__(self):
        return (
            f"Analysis Summary:\n"
            f"- Main Effects: \n\t" + "\n\t".join(str(var) for var in self.main_effects) + "\n"
            f"- Interaction Effects: \n\t" + "\n\t".join(str(var) for var in self.interaction_effects) + "\n"
            f"- Time-Varying Effects: \n\t" + "\n\t".join(str(var) for var in self.time_varying_effects) + "\n"
            f"- Within-Subjects Comparisons: \n\t" + "\n\t".join(str(var) for var in self.ws_comparisons) + "\n"
        )
    

def compare(design1: Design, design2: Design):
    """
    Compares two designs and prints the effects that are testable in one but not the other.
    + indicates effects that are testable in design 1 but not design 2.
    - indicates effects that are testable in design 2 but not design 1.
    """

    analysis1 = Analysis(design1)
    analysis2 = Analysis(design2)

    # Compute differences and intersections
    design1_main_effects = analysis1.main_effects.difference(analysis2.main_effects)
    design2_main_effects = analysis2.main_effects.difference(analysis1.main_effects)
    shared_main_effects = analysis1.main_effects.intersection(analysis2.main_effects)

    design1_interaction_effects = analysis1.interaction_effects.difference(analysis2.interaction_effects)
    design2_interaction_effects = analysis2.interaction_effects.difference(analysis1.interaction_effects)
    shared_interaction_effects = analysis1.interaction_effects.intersection(analysis2.interaction_effects)

    design1_time_varying_effects = analysis1.time_varying_effects.difference(analysis2.time_varying_effects)
    design2_time_varying_effects = analysis2.time_varying_effects.difference(analysis1.time_varying_effects)

    design1_ws_comparisons = analysis1.ws_comparisons.difference(analysis2.ws_comparisons)
    design2_ws_comparisons = analysis2.ws_comparisons.difference(analysis1.ws_comparisons)

    print("Comparative Analysis")
    print("+ indicates effects testable in design 1 but not design 2.")
    print("- indicates effects testable in design 2 but not design 1.\n")

    # Main effects
    print("Main Effects:")
    for var in design1_main_effects:
        print(f"\t+ {var}")
    for var in design2_main_effects:
        print(f"\t- {var}")
    for var in shared_main_effects:
        # Check if the shared main effect is conditional on an interaction effect
        for interaction_var in design1_interaction_effects:
            if var in interaction_var.get_variables():
                print(f"\t+ {var} (under assumption of interaction with {interaction_var})")
        for interaction_var in design2_interaction_effects:
            if var in interaction_var.get_variables():
                print(f"\t- {var} (under assumption of interaction with {interaction_var})")

        for time_var in design1_time_varying_effects:
            if not time_var.is_multifact() and var in time_var.get_variables():
                print(f"\t+ {var} (under weaker assumption of no time-varying effect)")

        for time_var in design2_time_varying_effects:
            if not time_var.is_multifact() and var in time_var.get_variables():
                print(f"\t- {var} (under weaker assumption of no time-varying effect)")

    # Interaction effects
    print("\nInteraction Effects:")
    for var in design1_interaction_effects:
        print(f"\t+ {var}")
    for var in design2_interaction_effects:
        print(f"\t- {var}")
    # Check if shared interaction effects are conditional on time-varying effects
    for var in shared_interaction_effects:
        if var in design1_time_varying_effects:
            print(f"\t+ {var} (under weaker assumption of no time-varying effect of {var})")
        if var in design2_time_varying_effects:
            print(f"\t- {var} (under weaker assumption of no time-varying effect of {var})")

    # Time-varying effects
    print("\nTime-Varying Effects:")
    for var in design1_time_varying_effects:
        print(f"\t+ {var}")
    for var in design2_time_varying_effects:
        print(f"\t- {var}")

    # Within-subjects comparisons
    print("\nWithin-Subjects Comparisons:")
    for var in design1_ws_comparisons:
        print(f"\tDesign 1 requires fewer participants than Design 2 to estimate the effect of {var}")
    for var in design2_ws_comparisons:
        print(f"\tDesign 2 requires fewer participants than Design 1 to estimate the effect of {var}")